from datetime import datetime

from openpyxl.styles import Alignment, Font
from openpyxl.worksheet.worksheet import Worksheet

from excel_exporter.configuration.cell_format import CellFormat


def write_update_time(
    ws: Worksheet,
    update_time: datetime,
    format: CellFormat,
    message: str = 'Updated at',
):
    cell = ws.cell(row=1, column=2)
    # Write update time
    cell.value = f'{message} {update_time.strftime("%Y-%m-%d")}'
    # Format update time cell
    alignment = Alignment(
        horizontal=format.horizontal_alignment,
        vertical=format.vertical_alignment,
        wrap_text=format.line_break,
    )
    font = Font(size=format.font_size)
    cell.alignment = alignment
    cell.font = font



def from_list_of_dicts(records_list):
    """Converts a list of dicts into a list of dicts"""
    return {
        col: [record[col] for record in records_list]
        for col in records_list[0]
    }


def from_pandas_dataframe(df):
    """Converts a pandas dataframe into a list of dicts"""
    return df.to_dict(orient='list')



from typing import Dict, Tuple

from excel_exporter.configuration.group import Group
from excel_exporter.configuration.column import Column
from excel_exporter.configuration.cell_format import CellFormat


def indexes(element, my_list):
    first = my_list.index(element) + 1
    last = len(my_list) - my_list[::-1].index(element)
    return first, last


class Sheet:
    def __init__(
        self,
        sheet_name: str,
        update_date_format: CellFormat,
        groups: Dict[str, Group],
        columns: Dict[str, Column],
    ):
        self.sheet_name = sheet_name
        self.update_date_format = update_date_format
        self.groups = groups
        self.columns = columns

    def groups_limits(self) -> Dict[str, Tuple]:
        cols_groups = [col.group.group_name for col in self.columns.values()]
        groups_names = list(dict.fromkeys(cols_groups))
        return {
            group_name: indexes(group_name, cols_groups)
            for group_name in groups_names
        }

    def __repr__(self):
        s = []
        s.append(f'Sheet()')
        s.append(f'sheet_name: {self.sheet_name}')
        s.append(f'{len(self.groups)} groups')
        s.append(f'{len(self.columns)} columns')
        return '\n'.join(s)



from typing import Dict, List

from openpyxl.styles import Alignment, Font
from openpyxl.worksheet.worksheet import Worksheet

from excel_exporter.configuration.column import Column as ColumnConfiguration
from excel_exporter.configuration.cell_format import CellFormat


def write_columns_data(
    ws: Worksheet,
    ws_data: Dict[str, List],
    columns_config: Dict[str, ColumnConfiguration],
    start_col: int = 2,
):
    # print(ws_data.keys())
    for col_number, col_config in enumerate(
        columns_config.values(), start=start_col
    ):
        col_data = ws_data[col_config.variable_name]
        format = col_config.cell_format
        write_column(ws, col_number, col_data, format)


def write_column(
    ws: Worksheet,
    col_number: int,
    col_data: List,
    format: CellFormat,
    start_row: int = 4,
):
    alignment = Alignment(
        horizontal=format.horizontal_alignment,
        vertical=format.vertical_alignment,
        wrap_text=format.line_break,
    )
    font = Font(size=format.font_size)
    url_font = Font(size=format.font_size, underline='single', color='0563C1')
    for row_number, value in enumerate(col_data, start=start_row):
        if isinstance(value, list) or isinstance(value, tuple):
            value, url = value[0], value[1]
        else:
            url = None
        cell = ws.cell(row=row_number, column=col_number)
        cell.value = value
        cell.alignment = alignment
        if not url:
            cell.font = font
        else:
            cell.hyperlink = url
            cell.font = url_font



from excel_exporter.configuration.cell_format import CellFormat
from excel_exporter.configuration.group import Group


class Column:
    def __init__(
        self,
        column_name: str,
        variable_name: str,
        cell_format: CellFormat,
        group: Group,
        column_width: int,
    ):
        self.column_name = column_name
        self.variable_name = variable_name
        self.cell_format = cell_format
        self.group = group
        self.column_width = column_width

    def __repr__(self):
        p = []
        p.append(f'column_name={self.column_name}')
        p.append(f'variable_name={self.variable_name}')
        p.append(f'cell_format={self.cell_format}')
        p.append(f'group={self.group}')
        p.append(f'column_width={self.column_width}')
        parameters = '\n'.join(p)
        return f'Column({parameters})'



from typing import Dict
import yaml


from excel_exporter.configuration.are_strings_grouped import (
    are_strings_grouped,
)
from excel_exporter.configuration.cell_format import CellFormat
from excel_exporter.configuration.column import Column
from excel_exporter.configuration.excel_configuration import ExcelConfiguration
from excel_exporter.configuration.group import Group
from excel_exporter.configuration.sheet import Sheet


def parse_yaml(yaml_data: Dict) -> ExcelConfiguration:
    # 1. File name
    file_name = yaml_data['file_name']
    # 2. Cell Formats
    cell_formats = {}
    for cell_format_data in yaml_data['cell_formats']:
        cell_format = CellFormat(
            cell_format_data['format_name'],
            cell_format_data['font_size'],
            cell_format_data['horizontal_alignment'],
            cell_format_data['vertical_alignment'],
            cell_format_data['line_break'],
        )
        cell_formats[cell_format.format_name] = cell_format
    # 3. Sheets
    sheets = {}
    for sheet_data in yaml_data['sheets']:
        # 3,1. In each sheet, groups
        groups = {}
        for group_data in sheet_data['groups']:
            group = Group(
                group_data['group_name'], group_data['background_color']
            )
            groups[group.group_name] = group
        # 3.2 In each sheet, columns
        columns = {}
        for col_data in sheet_data['columns']:
            # Each column references to a previously loaded cell formart
            cell_format = cell_formats.get(col_data['cell_format'], None)
            if not cell_format:
                raise ValueError(
                    f"Cell format {col_data['cell_format']} not found"
                )
            # Each column references to a previously loaded group
            group = groups.get(col_data['group'], None)
            if not group:
                raise ValueError(f"Group {col_data['group']} not found")
            # Finish creating the column
            column = Column(
                col_data['column_name'],
                col_data['variable_name'],
                cell_format,
                group,
                col_data['column_width'],
            )
            columns[column.column_name] = column
        # Verify if columns are really grouped
        cols_groups = [col.group.group_name for col in columns.values()]
        if not are_strings_grouped(cols_groups):
            raise ValueError(f'Columns are not grouped: \n{cols_groups}')
        # Update date reference to a previously load cell format
        update_date_format = cell_formats.get(
            sheet_data['update_date_format'], None
        )
        if not update_date_format:
            raise ValueError(
                f"Cell format {sheet_data['update_date_format']} not found"
            )
        # Finish creating the sheet
        sheet = Sheet(
            sheet_data['sheet_name'], update_date_format, groups, columns
        )
        sheets[sheet_data['sheet_name']] = sheet
    # Finish creating the configuration
    return ExcelConfiguration(file_name, cell_formats, sheets)


def load_config(file_path: str) -> ExcelConfiguration:
    with open(file_path, 'r') as file:
        yaml_data = yaml.safe_load(file)
    return parse_yaml(yaml_data)



import re

from openpyxl.utils import get_column_letter, column_index_from_string
from openpyxl.worksheet.table import Table, TableStyleInfo
from openpyxl.worksheet.worksheet import Worksheet

from excel_exporter.configuration.sheet import Sheet as SheetConfiguration


def make_table_name(sheet_name: str) -> str:
    table_name = sheet_name.replace(' ', '_').lower()
    table_name = re.sub(r'[^a-zA-Z0-9_]+', '', table_name)
    return table_name


def apply_sheet_configs(ws: Worksheet, ws_config: SheetConfiguration):
    # Sheet name
    ws.title = ws_config.sheet_name
    # Format Table
    ref = f'B3:{get_column_letter(ws.max_column)}{str(ws.max_row)}'
    table_name = make_table_name(ws_config.sheet_name)
    table = Table(displayName=table_name, ref=ref)
    style = TableStyleInfo(
        name='TableStyleMedium1',
        showFirstColumn=False,
        showLastColumn=False,
        showRowStripes=True,
        showColumnStripes=False,
    )
    table.tableStyleInfo = style
    ws.add_table(table)
    # Adjust border columns widths
    left_border_col = get_column_letter(ws.min_column - 1)
    right_border_col = get_column_letter(ws.max_column + 1)
    ws.column_dimensions[left_border_col].width = 2
    ws.column_dimensions[right_border_col].width = 2
    # Hide gridlines
    ws.sheet_view.showGridLines = False
    # Hide unused columns:
    hidden_begin = ws.max_column + 2
    hidden_end = column_index_from_string('XFD') + 1
    for col_number in range(hidden_begin, hidden_end):
        col_letter = get_column_letter(col_number)
        ws.column_dimensions[col_letter].hidden = True



class Group:
    def __init__(self, group_name: str, background_color: str):
        self.group_name = group_name
        self.background_color = background_color

    def __repr__(self):
        p = []
        p.append(f'group_name={self.group_name}')
        p.append(f'background_color={self.background_color}')
        parameters = ', '.join(p)
        return f'Group({parameters})'



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['excel_exporter',
 'excel_exporter.configuration',
 'excel_exporter.exporter',
 'excel_exporter.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'openpyxl>=3.1.1,<4.0.0']

entry_points = \
{'console_scripts': ['example = example:example']}

setup_kwargs = {
    'name': 'excel-exporter-bms',
    'version': '0.1.2',
    'description': 'An excel exporter based on YAML configuration files',
    'long_description': '# Excel Exporter\n\n## Motivation\n\nFor many years, I have been creating excel spreadsheets with a similar look and feel. My template consists of data organized into excel tables with similar categories having the same background colors and a merged cell with a category title above them (outside the table). The background colors are dark(ish) to provide contrast with a bold white font. I use blank columns before and after the table as borders and hide grid lines and unused columns.\n\n![Image showing an example excel sheet as described above (check homepage)](images/img01_spreadsheet.png)\n\nI have been manually creating these spreadsheets in Excel, but when I needed to update one of them every week, I decided to create a python script to automate the process. The script worked well, but it was very hard to maintain. If I needed to add or remove a column, I would have to make a lot of changes, and if I needed to adapt the script to another dataset, I would have to almost start from scratch.\n\nI promised myself that I would make a cleaner version of the excel exporter as soon as I had time. The original script was written using the pandas and xlswriter libraries, where I put the data on the sheet using pandas and formatted everything I needed using xlswriter. However, recently I needed to add hyperlinks to some cells in a particular column, and I was unable to do so with my current approach. So, I decided to completely rewrite my script to make it easier to adapt to new uses and allow for the possibility of adding hyperlinks to cells.\n\nI don\'t intend to maintain this repository. However, I believe that the new version makes it very easy to create spreadsheets with my template and may be useful to others. Additionally, I think that the code is organized well enough to be adapted to other templates or used as a reference on how to organize a script to generate excel spreadsheets with complex formatting.\n\n\n## The new approach\n\nThe Excel Exporter was created utilizing openpyxl and has the capability to export multiple sheets into a single Excel file.\n\nThe data being exported is sourced from dictionaries (one for each sheet). In cases where the data contains associated URLs, the exporter will automatically convert them into hyperlinks within the Excel file.\n\nThe following animation shows how the data is organized:\n   \n![Gif animation showing the structure of data (check homepage)](images/img02_data.gif)\n\n Configurations for columns, groups, sheets, and the workbook as a whole are stored in a YAML file.\n - Workbook: File name, Cell formats and Sheets.\n - Cell Formats: Font size, Horizontal alignment, Vertical alignment and Line break\n - Sheets: Sheet Name, Group(s), and Column(s)\n - Groups: Name and Background Color\n - Columns: Order, Title (to be display in Excel), Variable Name (in the python dictionary), Cell format (previouly defined), Group it belongs to (previously defined) and  Column Width\n \n The following animation shows how the YAML file is organized:\n\n![Gif animation showing the structure of the YAML file (check homepage)](images/img03_yaml.gif)\n\n## How to install\n\nJust use pip\n\n```\npip install excel-exporter-bms\n```\n\nor poetry\n\n```\npoestry add excel-exporter-bms\n```\n\n## How to use\n\nTo run this tutorial you should download [example.py](https://github.com/brunomsantiago/excel_exporter/blob/master/example.py) and [example_config](https://github.com/brunomsantiago/excel_exporter/blob/master/example_config.yaml). It is expected to generate a file like [example_output.xlsx](https://github.com/brunomsantiago/excel_exporter/raw/master/example_output.xlsx)\n\n### 1. Prepare the Data\nIt\'s necessary to properly prepare your data before using the exporter.\nThe data must be a list of dictionary of lists, where each dictionary represents a sheet in the excel file.\nInside the dictionary each key represents a column and its corresponding list contains the column data. Each element in the list represents a cell value.\nIf a hyperlink is required, the cell value should be represented as a tuple, with the first element being the cell value and the second element being the URL for the hyperlink.\n```python\ndata = [\n    {\n        "employee_id": [(1001, \'https://employee.my_company.com/id/1001\'),\n                        (1002, \'https://employee.my_company.com/id/1002\')],\n        "employee_name": ["John Doe", "Jane Doe"],\n        "department": ["Sales South Area", "Sales West Area"],\n        "job_title": ["Manager", "Associate"]\n    },\n    {\n        "employee_id": [1001, 1002],\n        "sales_current_year": [100000, 120000],\n        "sales_last_year": [90000, 110000]\n    }\n]\n```\n\n- If your data originally comes from a list of dictionaries, you can use the `utils` module to convert it.\n```python\nfrom utils.data_conversion import from_list_of_dicts\n\nsheet1_data = from_list_of_dicts(sheet1_list_of_dicts)\ndata = [sheet1_data]\n```\n- If your data is in a pandas dataframe, a similar conversion process applies.\n```python\nfrom utils.data_conversion import from_pandas_dataframe\n\nsheet1_data = from_pandas_dataframe(sheet1_df)\ndata = [sheet1_data]\n```\n### 2. Prepare the YAML\nConfigure the YAML based on the example described on the previous section\n\n\n### 3. Load the YAML\nBefore using the exporter, you\'ll need to load the YAML configuration file that you\'ll be using to export your data.\n```python\nfrom configuration.load_config import load_config\n\nconfig = load_config(\'example_config.yaml\')\n```\n\n### 4. Retrieve the Current Date and Time\nThis exporter was created for periodically updated spreadsheets. So it also important to include the time and date of the dataset. These can be retrieved using datetime module.\n```python\nfrom datetime import datetime\n\nupdate_time = datetime.now()\n```\n\n### 5. Run the Exporter\nOnce your data is prepared and the YAML is loaded, you\'re ready to run the exporter. Simply pass your data, configuration and update time as arguments to the exporter, and the export process will begin. It will return a virtual file (BytesIO)\n```python\nfrom exporter.create_excel import export_excel\n\nvirtual_file = export_excel(data, config, update_time)\n```\n### 6. Save or upload the file\nOnce the export process is complete, you can choose to either save the virtual file as a real file on your local machine, or upload it to a remote server using an API. Here is an example of how to save the file locally:\n```python\nwith open(config.file_name, \'wb\') as file:\n    file.write(virtual_file.getbuffer())\n```',
    'author': 'Bruno Santiago',
    'author_email': 'brunomsantiago@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/brunomsantiago/excel_exporter',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
